use "ARITHMETIC_MACHINE.sig";
(* This machine has a simpler instruction set. *)
structure SimplerMachine :> ARITHMETIC_MACHINE =
struct
  datatype exp = Constant of int
               | Negate of exp
               | Add of exp * exp
               | Subtract of exp * exp
               | Multiply of exp * exp
  
  (* Notice we don't have Neg or Minus! *)
  datatype oper = Const of int
                | Plus
                | Times

  type code = oper list

  (* Instead we elaborate them away. *)
  fun elaborate e =
    case e of
      Negate e => Multiply (Constant ~1, elaborate e)
    | Subtract (e1, e2) => elaborate (Add (e1, Negate e2))
    | e => e

  exception CannotCompile

  fun compile e =
    case elaborate e of (* we elaborate here *)
      Constant i => [Const i]
    | Add (e1, e2) => (compile e1) @ (compile e2) @ [Plus]
    | Multiply (e1, e2) => (compile e1) @ (compile e2) @ [Times]
    | _ => raise CannotCompile

  type state = int list
  exception BadInput;

  fun unary f stack =
    case stack of
      e::tail => (f e)::tail
    | _ => raise BadInput

  fun binary f stack =
    case stack of
      r::l::tail => (f (l, r))::tail
    | _ => raise BadInput

  fun step oper =
    case oper of
      Const i => (fn s => i::s)
    | Plus => binary (op+)
    | Times => binary (op*)

  type value = int
  fun extract s =
    case s of
      [v] => v
    | _ => raise BadInput
end
